package Network.Algorithm.FlowAugmentingDecreasingPathsAlgorithm;

import Network.Algorithm.*;
import Network.Representation.*;
import Network.Algorithm.SimplexAlgorithm.*;

/**
 * This class represents a flow augmenting and decrementing paths algorithm relevant to minimum cost flow network simplex
 * algorithm represented by class <code>SimplexAlgorithm</code>. The code is based on : Warren B. Powell, "A Review Of
 * Sensitivity Results for Linear Networks and a New Approximation to Reduce the Effects of Degeneracy", Transportation
 * Science, volume 23, no 4, 1989, 231-243. This algorithm works on results that have been previously found by a <code>
 * SimplexAlgorithm</code>. Use of this class is just like any other algorithm: set the target network of the algorithm by
 * its <code>setTargetNetwork</code> method and use its core methods which is namely <code>findFlowAugmentingPaths</code> and
 * <code>findFlowDecreasingPaths</code>. The crucial thing is that the <code>SimplexAlgorithm</code> that is passes as arguments
 * to these core methods should have the same target network as this <code>FlowAugmentingDecreasingPathsAlgorithm</code>.
 */

public class FlowAugmentingDecreasingPathsAlgorithm extends AlgorithmImplementation {

    /**
     * Stands for the <code>SimplexAlgorithm</code> on whose results this algorithm will work on
     */

    public SimplexAlgorithm baseSimplexAlgorithm;
    
    private static final Node mNode = new NodeImplementation();
    private static final Node minus1Node = new NodeImplementation();


    public FlowAugmentingDecreasingPathsAlgorithm () {
        baseSimplexAlgorithm = null;
    }


    /**
     * Returns the this algorithm's <code>FlowAugmentingPathsAlgNodeInfo</code> corresponding to a node.
     * @param   whichNodesInfo  the node whose <code>FlowAugmentingDecreasingPathsAlgorithm</code> is to be returned.
     * @return  <code>FlowAugmentingPathsAlgNodeInfo</code> corresponding to a node.
     */
    public final FlowAugmentingDecreasingPathsAlgNodeInfo getAlgNodeInfo(Node whichNodesInfo) {
        return (FlowAugmentingDecreasingPathsAlgNodeInfo) (super.getGeneralAlgNodeInfo(whichNodesInfo));
    }


    /**
     * Returns the this algorithm's <code>FlowAugmentingPathsAlgArcInfo</code> corresponding to an arc.
     * @param   whichArcsInfo  the arc whose <code>FlowAugmentingDecreasingPathsAlgorithm</code> is to be returned.
     * @return  <code>FlowAugmentingPathsAlgArcInfo</code> corresponding to an arc.
     */
    public final FlowAugmentingDecreasingPathsAlgArcInfo getAlgArcInfo(Arc whichArcsInfo) {
        return (FlowAugmentingDecreasingPathsAlgArcInfo) (super.getGeneralAlgArcInfo(whichArcsInfo));
    }

    public final NodeAlgInfo getInstanceOfNodeAlgInfo() {
        FlowAugmentingDecreasingPathsAlgNodeInfo dummyAlgNodeInfo=AlgorithmImplementation.nodeInfoFactory.makeFlowInfo();
        return dummyAlgNodeInfo;
    }

    public final ArcAlgInfo getInstanceOfArcAlgInfo() {
        FlowAugmentingDecreasingPathsAlgArcInfo dummyAlgArcInfo=AlgorithmImplementation.arcInfoFactory.makeFlowInfo();
        return dummyAlgArcInfo;
    }

    /**
     * Finds the flow augmenting paths from each node to a given sink node. Flow augmenting paths are found by utilizing the results
     * that has been previously found by a <code>SimplexAlgorithm</code>. So the user has to specify which <code>SimplexAlgorithm</code>'s
     * results are to be used. Also the sink node which is the node to which the flow augmenting paths are to be found has to be specified.
     * In order to avoid cycling of the algorithm because of round off errors, the user has to specify whether the costs on the network are
     * doubles or just integers. Normally this method should NOT be used by users. Its main purpose is to provide a general functional
     * procedure that is to be used by other <code>findFlowAugmentingPaths</code> methods. If it is used, it should be kept in mind that
     * if <code>useDoubles</code> (which shows whether the costs on the network are doubles or not) parameter is given to be false then
     * <code>epsilon</code> (which shows the precision of comparison of doubles) parameter is totally ignored.
     * @param   baseSimplexAlgorithm_   the <code>SimplexAlgorithm</code> whose results are going to be used by this algorithm.
     * @param   sinkNode    the node to which all the flow augmenting paths are to be calculated.
     * @param   bigM        large cost needed to initialize the algorithm.
     * @param   useDoubles  <code>true</code> if the costs on arcs are doubles, <code>false</code> if the costs on arcs are integers.
     * @param   epsilon     the precision to be used in the comparison of flow augmenting duals.
     */
    protected void findFlowAugmentingPaths (SimplexAlgorithm baseSimplexAlgorithm_, Node sinkNode, double bigM , boolean useDoubles, double epsilon) {
        if (baseSimplexAlgorithm_ == null ) {
            throw new InvalidParameterError("can not find flow augmenting paths: base simplex algorithm must be non null");
        }
        else if (baseSimplexAlgorithm_.getTargetNetwork() == null) {
            throw new InvalidParameterError("can not find flow augmenting paths: base simplex algorithm must be directed to a network");
        }
        else if (targetNetwork == null) {
            throw new InvalidParameterError("can not find flow augmenting paths: target network must be non null");
        }
        else if (targetNetwork != baseSimplexAlgorithm_.getTargetNetwork()) {
            throw new InvalidParameterError("can not find flow augmenting paths: base simplex algorithm has a different target network");
        }
        else if (!(baseSimplexAlgorithm_.getTargetNetwork().getNodeList().contains(sinkNode))) {
            throw new InvalidParameterError("can not find flow augmenting paths: sink node is not in the target network");
        }

        else {
            Node last;
            Node n;
            Node m;

            baseSimplexAlgorithm=baseSimplexAlgorithm_;
            NodeList cursor=baseSimplexAlgorithm.getTargetNetwork().getNodeList();
            while (cursor!=null) {
                FlowAugmentingDecreasingPathsAlgNodeInfo myNodeInfo=getAlgNodeInfo(cursor.elem);
                myNodeInfo.flowAugmentingDual=bigM;
                myNodeInfo.flowAugmentingPathSuccessorNode=null;
                myNodeInfo.flowAugmentingPathSuccessorArc=null;
                myNodeInfo.flowAugmentingPathList=null;
                cursor=cursor.next;
            }

            getAlgNodeInfo(sinkNode).flowAugmentingDual=0;
            getAlgNodeInfo(sinkNode).flowAugmentingPathList=mNode;
            n=sinkNode;
            last=sinkNode;

            boolean shouldContinue=false;

            do {
                ArcList arcCursor=n.getInBoundArcs();
                while (arcCursor!=null) {
                    Arc myArc=arcCursor.elem;
                    Node i=myArc.getTailNode();
                    if (baseSimplexAlgorithm.getAlgArcInfo(myArc).getFlow()<myArc.getUpperBound()) {
                        FlowAugmentingDecreasingPathsAlgNodeInfo iNodeInfo = getAlgNodeInfo(i);
                        if ( ( ((getAlgNodeInfo(n).flowAugmentingDual+myArc.getCost()) < (iNodeInfo.flowAugmentingDual)) && (!useDoubles) ) || ( ((getAlgNodeInfo(n).flowAugmentingDual+myArc.getCost()) < (iNodeInfo.flowAugmentingDual-epsilon)) && (useDoubles) ) ){

                            iNodeInfo.flowAugmentingDual = (getAlgNodeInfo(n).flowAugmentingDual+myArc.getCost());
                            iNodeInfo.flowAugmentingPathSuccessorNode = n;
                            iNodeInfo.flowAugmentingPathSuccessorArc = myArc;

                            if ( iNodeInfo.flowAugmentingPathList == null ) {
                                getAlgNodeInfo(last).flowAugmentingPathList = i;
                                last = i;
                                iNodeInfo.flowAugmentingPathList = mNode;
                            }

                            else if ( iNodeInfo.flowAugmentingPathList == minus1Node ) {
                                iNodeInfo.flowAugmentingPathList = getAlgNodeInfo(n).flowAugmentingPathList;
                                getAlgNodeInfo(n).flowAugmentingPathList = i;
                            }
                        }
                    }
                    arcCursor=arcCursor.next;
                }

                arcCursor=n.getOutBoundArcs();
                while (arcCursor!=null) {
                    Arc myArc=arcCursor.elem;
                    Node j=myArc.getHeadNode();
                    if (baseSimplexAlgorithm.getAlgArcInfo(myArc).getFlow() > 0 ) {
                        FlowAugmentingDecreasingPathsAlgNodeInfo jNodeInfo = getAlgNodeInfo(j);
                        if ( ( ((getAlgNodeInfo(n).flowAugmentingDual - myArc.getCost()) < (jNodeInfo.flowAugmentingDual)) && (!useDoubles) ) || ( ((getAlgNodeInfo(n).flowAugmentingDual - myArc.getCost()) < (jNodeInfo.flowAugmentingDual-epsilon)) && (useDoubles) ) ){

                            jNodeInfo.flowAugmentingDual = (getAlgNodeInfo(n).flowAugmentingDual - myArc.getCost());
                            jNodeInfo.flowAugmentingPathSuccessorNode = n;
                            jNodeInfo.flowAugmentingPathSuccessorArc = myArc;

                            if ( jNodeInfo.flowAugmentingPathList == null ) {
                                getAlgNodeInfo(last).flowAugmentingPathList = j;
                                last = j;
                                jNodeInfo.flowAugmentingPathList = mNode;
                            }

                            else if ( jNodeInfo.flowAugmentingPathList == minus1Node ) {
                                jNodeInfo.flowAugmentingPathList = getAlgNodeInfo(n).flowAugmentingPathList;
                                getAlgNodeInfo(n).flowAugmentingPathList = j;
                            }
                        }
                    }
                    arcCursor=arcCursor.next;
                }

                m = n;
                shouldContinue=false;
                if ( getAlgNodeInfo(n).flowAugmentingPathList != mNode ) {
                    shouldContinue=true;
                    n = getAlgNodeInfo(n).flowAugmentingPathList;
                    getAlgNodeInfo(m).flowAugmentingPathList = minus1Node;
                }
            }
            while (shouldContinue);
        }
    }

    /**
     * Finds the flow augmenting paths from each node to a given sink node for a network in which all arc costs are integers.
     * @param   baseSimplexAlgorithm_   the <code>SimplexAlgorithm</code> whose results are going to be used by this algorithm.
     * @param   sinkNode    the node to which all the flow augmenting paths are to be calculated.
     * @param   bigM        large cost needed to initialize the algorithm.
     */
    public void findFlowAugmentingPaths (SimplexAlgorithm baseSimplexAlgorithm_, Node sinkNode, double bigM) {
        findFlowAugmentingPaths (baseSimplexAlgorithm_, sinkNode , bigM, false, 0);
    }

    /**
     * Finds the flow augmenting paths from each node to a given sink node for a network in which all arc costs are doubles.
     * @param   baseSimplexAlgorithm_   the <code>SimplexAlgorithm</code> whose results are going to be used by this algorithm.
     * @param   sinkNode    the node to which all the flow augmenting paths are to be calculated.
     * @param   bigM        large cost needed to initialize the algorithm.
     * @param   epsilon     the precision to be used in comparison of doubles.
     */
    public void findFlowAugmentingPaths (SimplexAlgorithm baseSimplexAlgorithm_, Node sinkNode, double bigM, double epsilon) {
        findFlowAugmentingPaths (baseSimplexAlgorithm_, sinkNode , bigM, true, epsilon);
    }




    /**
     * Finds the flow decreasing paths from each node to a given sink node. Flow decreasing paths are found by utilizing the results
     * that has been previously found by a <code>SimplexAlgorithm</code>. So the user has to specify which <code>SimplexAlgorithm</code>'s
     * results are to be used. Also the sink node which is the node to which the flow decreasing paths are to be found has to be specified.
     * In order to avoid cycling of the algorithm because of round off errors, the user has to specify whether the costs on the network are
     * doubles or just integers. Normally this method should NOT be used by users. Its main purpose is to provide a general functional
     * procedure that is to be used by other <code>findFlowDecreasingPaths</code> methods. If it is used, it should be kept in mind that
     * if <code>useDoubles</code> (which shows whether the costs on the network are doubles or not) parameter is given to be false then
     * <code>epsilon</code> (which shows the precision of comparison of doubles) parameter is totally ignored.
     * @param   baseSimplexAlgorithm_   the <code>SimplexAlgorithm</code> whose results are going to be used by this algorithm.
     * @param   sinkNode    the node to which all the flow decreasing paths are to be calculated.
     * @param   bigM        large cost needed to initialize the algorithm.
     * @param   useDoubles  <code>true</code> if the costs on arcs are doubles, <code>false</code> if the costs on arcs are integers.
     * @param   epsilon     the precision to be used in the comparison of flow Decreasing duals.
     */
    protected void findFlowDecreasingPaths (SimplexAlgorithm baseSimplexAlgorithm_, Node sinkNode, double bigM, boolean useDoubles, double epsilon) {
        if (baseSimplexAlgorithm_ == null ) {
            throw new InvalidParameterError("can not find flow decreasing paths: base simplex algorithm must be non null");
        }
        else if (baseSimplexAlgorithm_.getTargetNetwork() == null) {
            throw new InvalidParameterError("can not find flow decreasing paths: base simplex algorithm must be directed to a network");
        }
        else if (targetNetwork == null) {
            throw new InvalidParameterError("can not find flow decreasing paths: target network must be non null");
        }
        else if (targetNetwork != baseSimplexAlgorithm_.getTargetNetwork()) {
            throw new InvalidParameterError("can not find flow decreasing paths: base simplex algorithm has a different target network");
        }
        else if (!(baseSimplexAlgorithm_.getTargetNetwork().getNodeList().contains(sinkNode))) {
            throw new InvalidParameterError("can not find flow decreasing paths: sink node is not in the target network");
        }

        else {
            Node last;
            Node n;
            Node m;

            baseSimplexAlgorithm=baseSimplexAlgorithm_; //Added this line, J.A.S. 9/10/98 to avoid NullPointerException
            NodeList cursor=baseSimplexAlgorithm.getTargetNetwork().getNodeList();
            while (cursor!=null) {
                FlowAugmentingDecreasingPathsAlgNodeInfo myNodeInfo=getAlgNodeInfo(cursor.elem);
                myNodeInfo.flowDecreasingDual=bigM;
                myNodeInfo.flowDecreasingPathSuccessorNode=null;
                myNodeInfo.flowDecreasingPathSuccessorArc=null;
                myNodeInfo.flowDecreasingPathList=null;
                cursor=cursor.next;
            }

            getAlgNodeInfo(sinkNode).flowDecreasingDual=0;
            getAlgNodeInfo(sinkNode).flowDecreasingPathList=mNode;
            n=sinkNode;
            last=sinkNode;

            boolean shouldContinue=false;
            do {
                ArcList arcCursor=n.getInBoundArcs();
                while (arcCursor!=null) {
                    Arc myArc=arcCursor.elem;
                    Node i=myArc.getTailNode();
                    if (baseSimplexAlgorithm.getAlgArcInfo(myArc).getFlow()>0) {
                        FlowAugmentingDecreasingPathsAlgNodeInfo iNodeInfo = getAlgNodeInfo(i);
                        if ( ( ((getAlgNodeInfo(n).flowDecreasingDual - myArc.getCost()) < (iNodeInfo.flowDecreasingDual)) && (!useDoubles) ) || ( ((getAlgNodeInfo(n).flowDecreasingDual - myArc.getCost()) < (iNodeInfo.flowDecreasingDual-epsilon)) && (useDoubles) ) ){

                            iNodeInfo.flowDecreasingDual = (getAlgNodeInfo(n).flowDecreasingDual - myArc.getCost());
                            iNodeInfo.flowDecreasingPathSuccessorNode = n;
                            iNodeInfo.flowDecreasingPathSuccessorArc = myArc;

                            if ( iNodeInfo.flowDecreasingPathList == null ) {
                                getAlgNodeInfo(last).flowDecreasingPathList = i;
                                last = i;
                                iNodeInfo.flowDecreasingPathList = mNode;
                            }

                            else if ( iNodeInfo.flowDecreasingPathList == minus1Node ) {
                                iNodeInfo.flowDecreasingPathList = getAlgNodeInfo(n).flowDecreasingPathList;
                                getAlgNodeInfo(n).flowDecreasingPathList = i;
                            }
                        }
                    }
                    arcCursor=arcCursor.next;
                }

                arcCursor=n.getOutBoundArcs();
                while (arcCursor!=null) {
                    Arc myArc=arcCursor.elem;
                    Node j=myArc.getHeadNode();
                    if (baseSimplexAlgorithm.getAlgArcInfo(myArc).getFlow() < myArc.getUpperBound() ) {
                        FlowAugmentingDecreasingPathsAlgNodeInfo jNodeInfo = getAlgNodeInfo(j);
                        if ( ( ((getAlgNodeInfo(n).flowDecreasingDual + myArc.getCost()) < (jNodeInfo.flowDecreasingDual)) && (!useDoubles) ) || ( ((getAlgNodeInfo(n).flowDecreasingDual + myArc.getCost()) < (jNodeInfo.flowDecreasingDual-epsilon)) && (useDoubles) ) ){

                            jNodeInfo.flowDecreasingDual = (getAlgNodeInfo(n).flowDecreasingDual + myArc.getCost());
                            jNodeInfo.flowDecreasingPathSuccessorNode = n;
                            jNodeInfo.flowDecreasingPathSuccessorArc = myArc;

                            if ( jNodeInfo.flowDecreasingPathList == null ) {
                                getAlgNodeInfo(last).flowDecreasingPathList = j;
                                last = j;
                                jNodeInfo.flowDecreasingPathList = mNode;
                            }

                            else if ( jNodeInfo.flowDecreasingPathList == minus1Node ) {
                                jNodeInfo.flowDecreasingPathList = getAlgNodeInfo(n).flowDecreasingPathList;
                                getAlgNodeInfo(n).flowDecreasingPathList = j;
                            }
                        }
                    }
                    arcCursor=arcCursor.next;
                }

                m = n;
                shouldContinue=false;
                if ( getAlgNodeInfo(n).flowDecreasingPathList != mNode ) {
                    shouldContinue=true;
                    n = getAlgNodeInfo(n).flowDecreasingPathList;
                    getAlgNodeInfo(m).flowDecreasingPathList = minus1Node;
                }
            }
            while (shouldContinue);
        }
    }

    /**
     * Finds the flow decreasing paths from each node to a given sink node for a network in which all arc costs are doubles.
     * @param   baseSimplexAlgorithm_   the <code>SimplexAlgorithm</code> whose results are going to be used by this algorithm.
     * @param   sinkNode    the node to which all the flow decreasing paths are to be calculated.
     * @param   bigM        large cost needed to initialize the algorithm.
     */
    public void findFlowDecreasingPaths (SimplexAlgorithm baseSimplexAlgorithm_, Node sinkNode, double bigM) {
        findFlowDecreasingPaths (baseSimplexAlgorithm_, sinkNode , bigM, false, 0);
    }

    /**
     * Finds the flow decreasing paths from each node to a given sink node for a network in which all arc costs are doubles.
     * @param   baseSimplexAlgorithm_   the <code>SimplexAlgorithm</code> whose results are going to be used by this algorithm.
     * @param   sinkNode    the node to which all the flow decreasing paths are to be calculated.
     * @param   bigM        large cost needed to initialize the algorithm.
     * @param   epsilon     the precision to be used in comparison of doubles.
     */
    public void findFlowDecreasingPaths (SimplexAlgorithm baseSimplexAlgorithm_, Node sinkNode, double bigM, double epsilon) {
        findFlowDecreasingPaths (baseSimplexAlgorithm_, sinkNode , bigM, true, epsilon);
    }

    /**
     * Return the string representation of this object which is the string representation of the target network for this algorithm and
     * the string representation of <code>baseSimplexAlgorithm</code> of this class.
     */
    public String toString() {
        return targetNetwork.toString()+" "+baseSimplexAlgorithm.toString();
    }



}