package Network.Algorithm.SimplexAlgorithm;


import Network.Representation.*;
import Network.Algorithm.*;
import Network.Algorithm.TreeAlgorithm.*;

/** 
 * This class represents a minimum cost flow, network simplex algorithm. Simplex algorithm
 * has a lot to do with trees so this algorithm also has a tree algorithm associated with
 * it. So each simplex algorithm is infact composed of two algorithms: one tree algorithm and itself.
 * Mostly a tree algorithm's node information is crucial for network simplex and correctness of
 * tree_algorithm_node_information for any <code>TreeAlgorithm</code> that is relevant to a <code>SimplexAlgorithm</code>
 * is guaranteed after calling <code>solve</code>  method of a <code>SimplexAlgorithm</code> instance, 
 * but users should not rely on the correctness of tree_algorithm_arc_information
 * relevant to a <code>TreeAlgorithm</code> that corresponds to a <code>SimplexAlgorithm</code> since simplex algorithm does not
 * utilize tree information relevant to an arc. The general procedure to use <code>SimplexAlgorithm</code> is just like nay other
 * algorithm: set the target network (the algorithm information that has been previously transcribed to another network
 * by this algorithm may be erased only at this stage is desired. Otherwise, it will remain there forever.) and call
 * the core method which is <code>solve</code> in this algorithm.
 */

public class SimplexAlgorithm extends AlgorithmImplementation {
    
    //
    protected Node artificialNode;
    protected CandidateList candidateList;
    private int noOfPivots;
    private int noOfDegeneratePivots;
    private double objectiveValue;
    public TreeAlgorithm treeAlgorithm;
    
    //Should add a factory for SimplexAlgArcInfo and SimplexNodeArcInfo
    //J.A.S., 8/31/98
    
    public SimplexAlgorithm(){
        super();
        treeAlgorithm=new TreeAlgorithm();
        //
        artificialNode = new NodeImplementation();
    }
    
    
     /**
      * @see Network.Algorithm.AlgorithmImplementation#setTargetNetwork
      */
    public void setTargetNetwork(Network setNetwork, boolean removeOldAlgInfo) {
        treeAlgorithm.setTargetNetwork (setNetwork, removeOldAlgInfo);
        super.setTargetNetwork(setNetwork, removeOldAlgInfo);
    }
    
    
    
    /** 
     * Returns node information relevant to this simplex algorithm as <code>SimplexAlgNodeInfo</code>
     * @param   whichNodesInfo     the node whose simplex algorithm node information is to be returned
     */  
    
    public final SimplexAlgNodeInfo getAlgNodeInfo(Node whichNodesInfo){
        return (SimplexAlgNodeInfo)(super.getGeneralAlgNodeInfo(whichNodesInfo));
    }
    
    
    
    /** 
     * Returns arc information relevant to this simplex algorithm.
     * @param   whichArcsInfo     the arc whose simplex algorithm arc information is to be returned
     */  
    public final SimplexAlgArcInfo getAlgArcInfo(Arc whichArcsInfo){
        return (SimplexAlgArcInfo)(super.getGeneralAlgArcInfo(whichArcsInfo));
    }
    
    public final NodeAlgInfo getInstanceOfNodeAlgInfo(){
        SimplexAlgNodeInfo dummyAlgNodeInfo=AlgorithmImplementation.nodeInfoFactory.makeSimplexInfo();
        return dummyAlgNodeInfo;   
    }
    
    public final ArcAlgInfo getInstanceOfArcAlgInfo(){
        SimplexAlgArcInfo dummyAlgArcInfo=AlgorithmImplementation.arcInfoFactory.makeSimplexInfo();
        return dummyAlgArcInfo;
    }
    
    public final int getNumberOfPivots() {
        return noOfPivots;
    }
    
    public final int getNumberOfDegeneratePivots(){
        return noOfDegeneratePivots;
    }
    
    public final double getObjectiveValue() {
        return objectiveValue;
    }
    
    public final double getFlow(Arc whichArc) {
        return getAlgArcInfo(whichArc).getFlow();
    }
    
    public final double getDualVariable(Node whichNode) {
        return getAlgNodeInfo(whichNode).getDualVariable();
    }
    
    /** 
     * Calculates the dual variables for the simplex algorithm assuming that tree structure is
     * correct. First the root is found and the duals are calculated. The structure of the tree is
     * NOT checked each time this method is run. Therefore, user is responsible for the correctness of
     * tree structure. Tree structure may be checked with <code>checkTree</code> method in <code>TreeAlgorithm</code>
     *.
     *
     *@see Network.Algorithm.TreeAlgorithm#checkTree
     *@see Network.Algorihtm.TreeAlgorihtm#findRoot
     */
     
    public String calculateDuals(){
        Node rootNode;
        Node topNode;
        String result;
        Node bottomNode;
        Arc predLink;
        double cost;

        result="DUALS_OK";
        rootNode=treeAlgorithm.findRoot();
        if (rootNode==null)
            result="DUALS_NO_ROOT";
        if (result=="DUALS_OK"){
            getAlgNodeInfo(rootNode).dualVariable=0;
            bottomNode=treeAlgorithm.getAlgNodeInfo(rootNode).forwardThread;
            while (bottomNode!=rootNode){
                topNode=treeAlgorithm.getAlgNodeInfo(bottomNode).treePredecessor;
                predLink=treeAlgorithm.getAlgNodeInfo(bottomNode).predecessorLink;
                cost=predLink.getCost();
                if ( treeAlgorithm.getAlgNodeInfo(bottomNode).downPointer ) {
                    getAlgNodeInfo(bottomNode).dualVariable=getAlgNodeInfo(topNode).dualVariable + cost;
                }
                else {
                    getAlgNodeInfo(bottomNode).dualVariable=getAlgNodeInfo(topNode).dualVariable - cost;
                }
                bottomNode=treeAlgorithm.getAlgNodeInfo(bottomNode).forwardThread;
            }
        }
        return result;
    }


    
    /** 
     * Carries out the "bigM" start procedure for network simplex. Since this procedure requires creating so many
     * dummy arcs, an arcfactory may be used to produce these dummy arcs.
     * This method should normally NOT be called by the users, because simplex algorithm automatically calls it
     * initially.
     * @param   bigM    large cost that forces the flows on dummy arcs to zero
     * @param   factory arc factory to be used, may be null if no arc factory is to be used
     * @param   useFactory  specifies whether the arc factory is to be used or not, sould be false if
     * the factory is specified to be null
     */
    
    protected final void bigMStart(double bigM, ArcFactory factory, boolean useFactory){
        
        
        //
        //Node artificialNode=new NodeImplementation();
        //
        targetNetwork.addNode(artificialNode);
        
        SimplexAlgNodeInfo dummySimplexAlgNodeInfo;
        TreeNodeInfo dummyTreeNodeInfo;
        
        artificialNode.getNodeAlgorithmInfo().setSize( Math.max( getAlgorithmOrder(), artificialNode.getNodeAlgorithmInfo().capacity()) );
        dummySimplexAlgNodeInfo=AlgorithmImplementation.nodeInfoFactory.makeSimplexInfo();
        dummyTreeNodeInfo=AlgorithmImplementation.nodeInfoFactory.makeTreeInfo();
        dummyTreeNodeInfo.depth=0;
        dummyTreeNodeInfo.treePredecessor=null;
        
        NodeList lastNode;
        lastNode=targetNetwork.getNodeListEnd();
        
        dummyTreeNodeInfo.reverseThread=lastNode.elem;
        dummyTreeNodeInfo.forwardThread=targetNetwork.getNodeList().elem;
        dummyTreeNodeInfo.predecessorLink=null;
        dummySimplexAlgNodeInfo.dualVariable=0;
        dummyTreeNodeInfo.downPointer=false;
        artificialNode.getNodeAlgorithmInfo().setElementAt(dummySimplexAlgNodeInfo,(getAlgorithmOrder()-1));
        artificialNode.getNodeAlgorithmInfo().setElementAt(dummyTreeNodeInfo,(treeAlgorithm.getAlgorithmOrder()-1));
        artificialNode.setSupply(0);
        
        
        NodeList nodeCursor=targetNetwork.getNodeList();
        ArcList arcCursor;
        Node myNode;
        Arc myArc;
        Arc dummyArc;
        SimplexAlgArcInfo dummySimplexAlgArcInfo;
        TreeArcInfo dummyTreeArcInfo;
        
        while (nodeCursor.elem!=artificialNode) {
            myNode=nodeCursor.elem;
            arcCursor=myNode.getOutBoundArcs();
            while  (arcCursor!=null) {
                myArc=arcCursor.elem;
                getAlgArcInfo(myArc).flow=0;
                arcCursor=arcCursor.next;
            }
            //Was > only
            if ((myNode.getSupply())>0) {
                getAlgNodeInfo(myNode).dualVariable=-bigM;
                treeAlgorithm.getAlgNodeInfo(myNode).downPointer=false;
                if (useFactory) {
                    dummyArc=factory.makeArc(bigM,1e15);
                    if ( dummyArc==null ) {dummyArc=new ArcImplementation(bigM,1e15);}
                }
                else {
                    dummyArc=new ArcImplementation(bigM,(1e15));
                   
                }
                targetNetwork.addArc(dummyArc,myNode,artificialNode);
                dummyArc.getArcAlgorithmInfo().setSize( Math.max(getAlgorithmOrder(), dummyArc.getArcAlgorithmInfo().capacity()) );
                dummySimplexAlgArcInfo=(SimplexAlgArcInfo)dummyArc.getArcAlgorithmInfo().elementAt(getAlgorithmOrder()-1);
                dummySimplexAlgArcInfo.flow=myNode.getSupply();
                
            }
            else{
                getAlgNodeInfo(myNode).dualVariable=bigM;
                treeAlgorithm.getAlgNodeInfo(myNode).downPointer=true;
                if (useFactory) {
                    dummyArc=factory.makeArc(bigM,1e15);
                    if ( dummyArc==null ) {dummyArc=new ArcImplementation(bigM,1e15);}
                }
                else {
                    dummyArc=new ArcImplementation(bigM,(1e15));
                }
                targetNetwork.addArc(dummyArc,artificialNode,myNode);
                dummyArc.getArcAlgorithmInfo().setSize( Math.max(getAlgorithmOrder(), dummyArc.getArcAlgorithmInfo().capacity()) );
                dummySimplexAlgArcInfo=(SimplexAlgArcInfo)dummyArc.getArcAlgorithmInfo().elementAt(getAlgorithmOrder()-1);
                dummySimplexAlgArcInfo.flow=-(myNode.getSupply());
                
            }
            TreeNodeInfo myNodeInfo = treeAlgorithm.getAlgNodeInfo(myNode);
            myNodeInfo.treePredecessor=artificialNode;
            myNodeInfo.depth=1;
            myNodeInfo.predecessorLink=dummyArc;
            NodeList previousOfNodeCursor;
            NodeList nextOfNodeCursor;
            previousOfNodeCursor=nodeCursor;
            nextOfNodeCursor=nodeCursor;
            if (previousOfNodeCursor.equals(targetNetwork.getNodeList())) {
                previousOfNodeCursor=targetNetwork.getNodeListEnd();
            }
            else {
                previousOfNodeCursor=previousOfNodeCursor.prev;
            }
            nextOfNodeCursor=nextOfNodeCursor.next;
            myNodeInfo.reverseThread=previousOfNodeCursor.elem;
            myNodeInfo.forwardThread=nextOfNodeCursor.elem;
            nodeCursor=nodeCursor.next;
        }
    }
    
    
    
    /**
     * Calculates the objective function value by simply going thru all arcs and multiplying the flows with costs
     * on all arcs and summing them up.
     */
    public double calculateObjective(){
        double result;
        NodeList nodeCursor=targetNetwork.getNodeList();
        ArcList arcCursor;
        result=0;
        while (nodeCursor!=null) {
           arcCursor=nodeCursor.elem.getOutBoundArcs();
           while (arcCursor!=null) {
            result+= arcCursor.elem.getCost() * getAlgArcInfo(arcCursor.elem).flow;
            arcCursor=arcCursor.next;
           }
           nodeCursor=nodeCursor.next;
        }
        return result;
    }
    
    
    /**
     * Checks whether the total demad is equal to total supply in the network.
     * @exception  InfeasibleException thrown if total supply is not equal to total demand
     */
     
    public final void checkDemandSupply() throws InfeasibleException{
        
        NodeList nodeCursor=targetNetwork.getNodeList();
        double sum=0;
        while (nodeCursor!=null) {
            sum+=nodeCursor.elem.getSupply();
            nodeCursor=nodeCursor.next;
        }
        if (sum!=0) throw new InfeasibleException("infeasible network: demand is not equal to supply");
        
    }
    
    
    /** 
     * Does the simplex iterations of the network simplex method. This procedure should normally NOT be called by users,
     * the way to solve simplex is by using the various <code>solve</code> methods.
     */
     
    protected void doSimplexIterations (double optimalityTolerance, int candidateListSize, int candidateRefillCycle, boolean isDoubleCosts) {
        Arc linkEnter;
        Node enterFrom;
        Node enterTo;
        double fmin0;
        double fmin1;
        double fmin2;
        double enterCost;
        double enteredCo;
        boolean moreFlag;
        Node k1Node;
        Node k2Node;
        int deep;
        Node tmpNode;
        Arc tmpArc;
        int i;
        Node blockNode0;
        Arc blockArc0;
        Node blockNode1=null;
        Arc blockArc1=null;
        Node blockNode2=null;
        Arc blockArc2=null;
        double dualAdj;
        boolean side1Flag;
        boolean treeFlag;
        int tmpInt;
        double tmpDouble;
        Node blFrNode;
        Node blToNode;
        boolean down1Flag;
        boolean down2Flag;
        int depthb;
        int depthc;
        Node nLastNode;
        candidateList=new ArrayCandidateList(candidateListSize, candidateRefillCycle, targetNetwork, this);
        objectiveValue=calculateObjective();
        noOfPivots=0;
        noOfDegeneratePivots=0;
        
        while ((linkEnter=candidateList.bestCandidate(optimalityTolerance,isDoubleCosts)) != null){
                
            noOfPivots++;
            //For debugging: System.err.println(noOfPivots + " : " + calculateObjective() + " " + linkEnter);
            enterFrom=linkEnter.getTailNode();
            enterTo=linkEnter.getHeadNode();
            fmin1=1e15;
            fmin2=1e15;
            enterCost=linkEnter.getCost();
            enteredCo=getAlgNodeInfo(enterFrom).dualVariable+enterCost-getAlgNodeInfo(enterTo).dualVariable;
            
            if (! (moreFlag=(enteredCo>0))){
                k2Node=enterTo;
                k1Node=enterFrom;
            }
            else{
                k1Node=enterTo;
                k2Node=enterFrom;
            }
            
            if ((deep=(treeAlgorithm.getAlgNodeInfo(k1Node).depth-treeAlgorithm.getAlgNodeInfo(k2Node).depth))<=0) {
                deep=-deep;
                tmpNode=k2Node;
                for (i=1;(i<=deep)&&(fmin2>0);i++){
                    tmpArc=treeAlgorithm.getAlgNodeInfo(tmpNode).predecessorLink;
                    if (treeAlgorithm.getAlgNodeInfo(tmpNode).downPointer) {
                        if (getAlgArcInfo(tmpArc).flow<fmin2) {
                            blockArc2=tmpArc;
                            blockNode2=tmpNode;
                            fmin2=getAlgArcInfo(tmpArc).flow;
                        }
                    }
                    else {
                        tmpDouble=tmpArc.getUpperBound()-getAlgArcInfo(tmpArc).flow;
                        if (tmpDouble<=fmin2) {
                            blockArc2=tmpArc;
                            blockNode2=tmpNode;
                            fmin2=tmpDouble;
                        }
                    }
                    tmpNode=treeAlgorithm.getAlgNodeInfo(tmpNode).treePredecessor;
                }
                blFrNode=k1Node;
                blToNode=tmpNode;
            }
            else {
                tmpNode=k1Node;
                for (i=1;(i<=deep) && (fmin1>0); i++){
                    tmpArc=treeAlgorithm.getAlgNodeInfo(tmpNode).predecessorLink;
                    if (!(treeAlgorithm.getAlgNodeInfo(tmpNode).downPointer)) {
                        if (getAlgArcInfo(tmpArc).flow <=fmin1) {
                            blockArc1=tmpArc;
                            blockNode1=tmpNode;
                            fmin1=getAlgArcInfo(tmpArc).flow;
                        }
                    }
                    else {
                        tmpDouble=tmpArc.getUpperBound()-getAlgArcInfo(tmpArc).flow;
                        if (tmpDouble<fmin1){
                            blockArc1=tmpArc;
                            blockNode1=tmpNode;
                            fmin1=tmpDouble;
                        }
                    }
                    tmpNode=treeAlgorithm.getAlgNodeInfo(tmpNode).treePredecessor;
                }
                blFrNode=tmpNode;
                blToNode=k2Node;
            }
            
            while ((blFrNode != blToNode) && (fmin1>0) && (fmin2>0)) {
                tmpArc=treeAlgorithm.getAlgNodeInfo(blToNode).predecessorLink;
                tmpDouble=getAlgArcInfo(tmpArc).flow;
                if (!(treeAlgorithm.getAlgNodeInfo(blToNode).downPointer)) {tmpDouble=tmpArc.getUpperBound()-tmpDouble;}
                if (tmpDouble<fmin2) {
                    blockArc2=tmpArc;
                    blockNode2=blToNode;
                    fmin2=tmpDouble;
                }
                
                blToNode=treeAlgorithm.getAlgNodeInfo(blToNode).treePredecessor;
                tmpArc=treeAlgorithm.getAlgNodeInfo(blFrNode).predecessorLink;
                tmpDouble=getAlgArcInfo(tmpArc).flow;
                if (treeAlgorithm.getAlgNodeInfo(blFrNode).downPointer) tmpDouble=tmpArc.getUpperBound()-tmpDouble;
                if (tmpDouble<=fmin1) {
                    blockArc1=tmpArc;
                    blockNode1=blFrNode;
                    fmin1=tmpDouble;
                }
                blFrNode=treeAlgorithm.getAlgNodeInfo(blFrNode).treePredecessor;
            }
            
            if (side1Flag=(fmin1<fmin2)) {
                fmin0=fmin1;
                blockNode0=blockNode1;
                blockArc0=blockArc1;
            }
            else{
                fmin0=fmin2;
                blockNode0=blockNode2;
                blockArc0=blockArc2;
            }
            
            if (moreFlag) {
                tmpDouble=getAlgArcInfo(linkEnter).flow;
            }
            else {
                tmpDouble=linkEnter.getUpperBound()-getAlgArcInfo(linkEnter).flow;
            }
            treeFlag=true;
            if (fmin0>tmpDouble) {
                fmin0=tmpDouble;
                treeFlag=false;
            }
            
            if (fmin0>0) {
                tmpNode=k1Node;
                while (tmpNode != blToNode) {
                    tmpArc=treeAlgorithm.getAlgNodeInfo(tmpNode).predecessorLink;
                    if (treeAlgorithm.getAlgNodeInfo(tmpNode).downPointer) {
                        getAlgArcInfo(tmpArc).flow+=fmin0;
                    }
                    else{
                        getAlgArcInfo(tmpArc).flow-=fmin0;
                    }
                    tmpNode=treeAlgorithm.getAlgNodeInfo(tmpNode).treePredecessor;
                }
                
                tmpNode=k2Node;
                while (tmpNode!=blToNode) {
                    tmpArc=treeAlgorithm.getAlgNodeInfo(tmpNode).predecessorLink;
                    if (treeAlgorithm.getAlgNodeInfo(tmpNode).downPointer) {
                        getAlgArcInfo(tmpArc).flow-=fmin0;
                    }
                    else {
                        getAlgArcInfo(tmpArc).flow+=fmin0;
                    }
                    tmpNode=treeAlgorithm.getAlgNodeInfo(tmpNode).treePredecessor;
                }
                
                if (moreFlag) fmin0=-fmin0;
                getAlgArcInfo(linkEnter).flow+=fmin0;
                objectiveValue+=enteredCo*fmin0;
            }
            else noOfDegeneratePivots++;
            
            if (treeFlag) {
                if (!side1Flag) {
                    blockNode1=k1Node;
                    blockNode2=k2Node;
                    down1Flag=treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer;
                    treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer=true;
                }
                else {
                    blockNode2=k1Node;
                    blockNode1=k2Node;
                    down1Flag=treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer;
                    treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer=false;
                    enterCost=-enterCost;
                }
                
                if (moreFlag) {
                    enterCost=-enterCost;
                    treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer=!(treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer);
                }
                
                dualAdj = getAlgNodeInfo(blockNode1).dualVariable + enterCost - getAlgNodeInfo(blockNode2).dualVariable;
                
                while (blockNode2!=null) {
                    depthb=treeAlgorithm.getAlgNodeInfo(blockNode2).depth;
                    depthc=treeAlgorithm.getAlgNodeInfo(blockNode1).depth+1-depthb;
                    
                    tmpNode=blockNode2;
                    do {
                        nLastNode=tmpNode;
                        getAlgNodeInfo(tmpNode).dualVariable+=dualAdj;
                        treeAlgorithm.getAlgNodeInfo(tmpNode).depth+=depthc;
                        tmpNode=treeAlgorithm.getAlgNodeInfo(tmpNode).forwardThread;
                    }while (treeAlgorithm.getAlgNodeInfo(tmpNode).depth>depthb);
                    
                    k1Node=treeAlgorithm.getAlgNodeInfo(nLastNode).forwardThread;
                    k2Node=treeAlgorithm.getAlgNodeInfo(blockNode2).reverseThread;
                    treeAlgorithm.getAlgNodeInfo(k2Node).forwardThread=k1Node;
                    treeAlgorithm.getAlgNodeInfo(k1Node).reverseThread=k2Node;
                    treeAlgorithm.getAlgNodeInfo(nLastNode).forwardThread=treeAlgorithm.getAlgNodeInfo(blockNode1).forwardThread;
                    treeAlgorithm.getAlgNodeInfo(blockNode1).forwardThread=blockNode2;
                    treeAlgorithm.getAlgNodeInfo(treeAlgorithm.getAlgNodeInfo(nLastNode).forwardThread).reverseThread=nLastNode;
                    treeAlgorithm.getAlgNodeInfo(blockNode2).reverseThread=blockNode1;
                    
                    blockArc1=treeAlgorithm.getAlgNodeInfo(blockNode2).predecessorLink;
                    k1Node=treeAlgorithm.getAlgNodeInfo(blockNode2).treePredecessor;
                    treeAlgorithm.getAlgNodeInfo(blockNode2).treePredecessor=blockNode1;
                    treeAlgorithm.getAlgNodeInfo(blockNode2).predecessorLink=linkEnter;
                    
                    if ((blockNode2!=blockNode0) && (blockNode1 != blockNode0)) {
                        blockNode1=blockNode2;
                        blockNode2=k1Node;
                        linkEnter=blockArc1;
                        
                        down2Flag=treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer;
                        treeAlgorithm.getAlgNodeInfo(blockNode2).downPointer= !down1Flag;
                        down1Flag=down2Flag;
                    }
                    else {
                        blockNode2=null;
                    }
                }
            }
            
        }
                    
    }
    
    
    
    /** 
     * Checks whether all the flows in and out of a node are all equal to 0. This is especially usefull at the end
     * of simplex algorithm to see whether all the flows in and out of dummy (artificial) node are all equal to 0. Normally users should
     * NOT need to access this method.
     */
    protected final boolean checkNodeInOutFlows(Node whichNode) {
        double totalInFlow=0;
        double totalOutFlow=0;
        ArcList arcCursor=whichNode.getOutBoundArcs();
        while (arcCursor!=null) {
            totalOutFlow += getAlgArcInfo(arcCursor.elem).flow;
            arcCursor=arcCursor.next;
        }
        arcCursor=whichNode.getInBoundArcs();
        while (arcCursor!=null) {
            totalInFlow += getAlgArcInfo(arcCursor.elem).flow;
            arcCursor=arcCursor.next;
        }
        return ((totalInFlow==0) && (totalInFlow==0));
    }
 
   
    /** 
     * Carries out network simplex. This particular form of <code> solve</code> procedure should normally NOT be used
     * by users or if used, it should be kept in mind that if <code>isDoubleCosts</code> is false, then <code>optimalityTolerance</code>
     * will totally be ignored.
     * @exception   InfeasibleException thrown if the problem is infeasible
     */
    public void solve (double bigM, double optimalityTolerance, int candidateListSize, int candidateListRefillCycle, boolean isDoubleCosts) throws InfeasibleException {
        if (targetNetwork == null ) {
            throw new InvalidOperationError("can't do simplex iterations: target network is null");
        }
        if (candidateListSize>targetNetwork.getNoOfArcs()) {
            throw new InvalidParameterError("can't do simplex iterations: candidate list size is too big");
        }
        if (candidateListSize<candidateListRefillCycle) {
            throw new InvalidParameterError("can't do simplex iterations: refill cycle for the candidate list is lager than its size");
        }
        
        checkDemandSupply();
        bigMStart(bigM, null, false);
        calculateDuals();
        doSimplexIterations(optimalityTolerance, candidateListSize, candidateListRefillCycle, isDoubleCosts);
        NodeList nodesEnd=getTargetNetwork().getNodeListEnd();
        if (! checkNodeInOutFlows(nodesEnd.elem) ) {throw new InfeasibleException ("infeasible network: can not find a feasible solution");}
        
        //Remove the artificial node--arcs not returned to factory
        
        //
        ArcList arcList = nodesEnd.elem.getOutBoundArcs();
        while ( arcList != null )
        {
            getTargetNetwork().removeArc( arcList.elem , null );
            arcList = arcList.next;
        }
        
        arcList = nodesEnd.elem.getInBoundArcs();
        while ( arcList != null )
        {
            getTargetNetwork().removeArc( arcList.elem , null );
            arcList = arcList.next;
        }
        //
        getTargetNetwork().removeNode(nodesEnd.elem, null, null);
        
        
        
        objectiveValue=calculateObjective();
        candidateList.cleanList();
    }
    
    /**
     * Carries out network simplex assuming that the costs are <code>double</code>s.
     * @param   bigM    high cost for dummy arcs in the initialization procedure
     * @param   optimalityTolerance optimality tolearnce to decide whether an arc is eligible to enter the base or not
     * @param   candidateListSize size of the candidate list, should always be less than the total number of arcs in the network
     * @param   candidateListRefillCycle    the number of arcs that are going to be used from candidate list before it is updated
     * @exception   InfeasibleException thrown if the problem is infeasible
     */
     
    public void solve (double bigM, double optimalityTolerance,int candidateListSize, int candidateListRefillCycle) throws InfeasibleException {
        solve(bigM, optimalityTolerance,candidateListSize,candidateListRefillCycle,true);
    }
    
   
   /**
    * Carries out network simplex assuming that all the costs are integers, so there is no need for an optimality tolerance
    * @exception   InfeasibleException thrown if the problem is infeasible
    */
    
    public void solve (double bigM, int candidateListSize, int candidateListRefillCycle) throws InfeasibleException {
        solve(bigM,1e-3,candidateListSize,candidateListRefillCycle,false);
    }
        
    
    
    
   /** 
     * Carries out network simplex. This particular form of <code> solve</code> procedure should normally NOT be used
     * by users or if used it should be kept in mind that if <code>isDoubleCosts</code> is false, then <code>optimalityTolerance</code>
     * will totally be ignored. This form of <code>solve</code> also utilizes an arc factory for the bigM start procedure
     * @exception   InfeasibleException thrown if the problem is infeasible
     */
   
   
    public void solve (double bigM, double optimalityTolerance, ArcFactory factory, int candidateListSize, int candidateListRefillCycle,boolean isDoubleCosts) throws InfeasibleException {
        
        if (targetNetwork == null ) {
            throw new InvalidOperationError("can't do simplex iterations: target network is null");
        }
        if (candidateListSize>targetNetwork.getNoOfArcs()) {
            throw new InvalidParameterError("can't do simplex iterations: candidate list size is too big");
        }
        if (candidateListSize<candidateListRefillCycle) {
            throw new InvalidParameterError("can't do simplex iterations: refill cycle for the candidate list is lager than its size");
        }
        Arc dummyArc;
        checkDemandSupply();
        bigMStart(bigM, factory, true);
        calculateDuals();
        doSimplexIterations(optimalityTolerance, candidateListSize, candidateListRefillCycle,isDoubleCosts);
        Node artificialNode=getTargetNetwork().getNodeListEnd().elem;
        if (! checkNodeInOutFlows(artificialNode) ) {throw new InfeasibleException ("infeasible network: can not find a feasible solution");}
        
        
        //Remove the artificial node and return arcs to factory
        ArcList arcCursor=artificialNode.getOutBoundArcs();
        while (arcCursor!=null) {
            dummyArc=arcCursor.elem;
            targetNetwork.removeArc(dummyArc, factory);
            arcCursor=arcCursor.next;
        }
        arcCursor=artificialNode.getInBoundArcs();
        while (arcCursor!=null) {
            dummyArc=arcCursor.elem;
            targetNetwork.removeArc(dummyArc, factory);
            arcCursor=arcCursor.next;
        }
        getTargetNetwork().removeNode(artificialNode, null, factory);
        
        objectiveValue=calculateObjective();

        candidateList.cleanList();
    }
    
    
    /**
     * Carries out network simplex assuming that the costs are <code>double</code>s and a non-null arc factory is to be used.
     * @param   bigM    high cost for dummy arcs in the initialization procedure
     * @param   optimalityTolerance optimality tolearnce to decide whether an arc is eligible to enter the base or not
     * @param factory   the non-null arc factory to be utilized in the initialization procedure
     * @param   candidateListSize size of the candidate list, should always be less than the total number of arcs in the network
     * @param   candidateListRefillCycle    the number of arcs that are going to be used from candidate list before it is updated
     * @exception   InfeasibleException thrown if the problem is infeasible
     */
    public void solve (double bigM, double optimalityTolerance, ArcFactory factory, int candidateListSize, int candidateListRefillCycle) throws InfeasibleException {
        solve (bigM, optimalityTolerance, factory, candidateListSize, candidateListRefillCycle, true);
    }
    /** 
     * Carries out network simplex assuming that the costs are integers and using a non-null arcfactory.
     * @exception   InfeasibleException thrown if the problem is infeasible
     */
     
    public void solve (double bigM, ArcFactory factory, int candidateListSize, int candidateListRefillCycle) throws InfeasibleException {
        solve (bigM, 1e-3, factory, candidateListSize, candidateListRefillCycle, false);
    }
    
    
    public void warmSolve (double bigM, ArcFactory factory, int candidateListSize, int candidateListRefillCycle ) 
       throws InfeasibleException
    {
      
        
        if (targetNetwork == null ) {
            throw new InvalidOperationError("can't do simplex iterations: target network is null");
        }
        if (candidateListSize>targetNetwork.getNoOfArcs()) {
            throw new InvalidParameterError("can't do simplex iterations: candidate list size is too big");
        }
        if (candidateListSize<candidateListRefillCycle) {
            throw new InvalidParameterError("can't do simplex iterations: refill cycle for the candidate list is lager than its size");
        }
        Arc dummyArc;
        checkDemandSupply();
        calculateDuals();
        doSimplexIterations( 1e-3 , candidateListSize, candidateListRefillCycle, false);
        candidateList.cleanList();
    }
    
    public void warmSolve (double bigM,  int candidateListSize, int candidateListRefillCycle ) 
       throws InfeasibleException
    {
      
        
        if (targetNetwork == null ) {
            throw new InvalidOperationError("can't do simplex iterations: target network is null");
        }
        if (candidateListSize>targetNetwork.getNoOfArcs()) {
            throw new InvalidParameterError("can't do simplex iterations: candidate list size is too big");
        }
        if (candidateListSize<candidateListRefillCycle) {
            throw new InvalidParameterError("can't do simplex iterations: refill cycle for the candidate list is lager than its size");
        }
        Arc dummyArc;
        checkDemandSupply();
        calculateDuals();
        doSimplexIterations( 1e-3 , candidateListSize, candidateListRefillCycle, false);
        candidateList.cleanList();
    }
    
    
    public void checkFlows() throws InfeasibleException
    {
         Node node;
         Arc arc;
         double totalFlow;
         ArcList arcList;
         
         NodeList nodeList = targetNetwork.getNodeList();
         
         while ( nodeList != null )
         {
            totalFlow = 0;
            arcList = nodeList.elem.getOutBoundArcs();
            while ( arcList != null )
            {
               totalFlow += getAlgArcInfo( arcList.elem ).flow;
               arcList = arcList.next;
            }
            
            arcList = nodeList.elem.getInBoundArcs();
            while ( arcList != null )
            {
               totalFlow = totalFlow - getAlgArcInfo( arcList.elem ).flow;
               arcList = arcList.next;
            }
            
            if ( totalFlow != nodeList.elem.getSupply() )
            {
              throw new InfeasibleException ( "unbalanced flow " + totalFlow + "<>" + nodeList.elem.getSupply() );
            }
            
            nodeList = nodeList.next;
         }
    }
    
    
    public void cleanFlows()
    {
         Node node;
         Arc arc;
         double totalFlow;
         ArcList arcList;

         NodeList nodeList = targetNetwork.getNodeList();

         while ( nodeList != null )
         {
            totalFlow = 0;
            arcList = nodeList.elem.getOutBoundArcs();
            while ( arcList != null )
            {
               getAlgArcInfo( arcList.elem ).flow = 0;
               arcList = arcList.next;
            }

            arcList = nodeList.elem.getInBoundArcs();
            while ( arcList != null )
            {
               getAlgArcInfo( arcList.elem ).flow = 0;
               arcList = arcList.next;
            }
            nodeList = nodeList.next;
         }
    }
    
    
    
    
    
    public String toString() {
        return Double.toString(objectiveValue)+" "+Integer.toString(noOfPivots)+" "+Integer.toString(noOfDegeneratePivots)+" "+super.toString();
    }
    
      
        
                
}

