package Network.Algorithm.TreeAlgorithm;


import Network.Representation.*;
import Network.Algorithm.*;

/** 
 * Trees are commonly used in many network algorithms. So basic operations that may be performed on
 * trees can be performed thru an instance of this class. For example <code>SimplexAlgorithm</code>
 * has a lot to do with trees so each simplex algorithm has a corresponding <code>TreeAlgorithm</code>.
 * Crucial information about trees are mostly transcribed in nodes, rather that arcs. So the correctness of
 * the tree_algorithm_node_information on nodes in guaranteed, however correctness of tree_algorithm_arc_information on
 * arcs is NOT guaranteed.
 */
public class TreeAlgorithm extends AlgorithmImplementation {
    
    //Should add a TreeArcInfo and TreeNodeInfo factory
    //J.A.S., 8/31/98
    
    public final TreeNodeInfo getAlgNodeInfo(Node whichNodesInfo){
        return (TreeNodeInfo)(super.getGeneralAlgNodeInfo(whichNodesInfo));
    }
    
    public final TreeArcInfo getAlgArcInfo(Arc whichArcsInfo){
        return (TreeArcInfo)(super.getGeneralAlgArcInfo(whichArcsInfo));
    }
    
    public final NodeAlgInfo getInstanceOfNodeAlgInfo(){
        TreeNodeInfo dummyAlgNodeInfo=AlgorithmImplementation.nodeInfoFactory.makeTreeInfo();
        return dummyAlgNodeInfo;   
    }
    
    public final ArcAlgInfo getInstanceOfArcAlgInfo(){
        TreeArcInfo dummyAlgArcInfo=AlgorithmImplementation.arcInfoFactory.makeTreeInfo();
        return dummyAlgArcInfo;
    }
    
     /**
      * Finds the root node of a tree, which is the node that has null <code>treePredecessor</code>.
      * This method does not chech for multiple roots. So the first node that has null <code>
      * treePredecessor is returned.
      */
     public final Node findRoot(){
        NodeList nodeCursor=targetNetwork.getNodeList();
        Node rootNode=null;
        Node myNode;
        while ((nodeCursor!=null) && (rootNode==null)) {
            myNode=nodeCursor.elem;
            if (getAlgNodeInfo(myNode).treePredecessor==null)
                rootNode=myNode;
            nodeCursor=nodeCursor.next;
        }
        return rootNode;
    }

    
    /**
     * Sets the the down_pointer for all nodes. Down_pointer shows whether each node is a head node for the arc
     * that connects it to subtree that contains the root node or a tail node for the arc that connects it to the
     * subtree that contains the root node. This method assumes that <code>treePredecessor</code> node_algorithm_fileds are
     * correct.
     */
     
     
    public final void setDownPointer(){ 
        Node predNode;
        NodeList nodeCursor=targetNetwork.getNodeList();
        Node myNode;
        while (nodeCursor!=null) {
            myNode=nodeCursor.elem;
            predNode=getAlgNodeInfo(myNode).treePredecessor;
            if (predNode!=null){
                if ( myNode==getAlgNodeInfo(myNode).predecessorLink.getTailNode() ) {
                    getAlgNodeInfo(myNode).downPointer=false;
                }
                else{
                    getAlgNodeInfo(myNode).downPointer=true;
                }
            }            
            nodeCursor=nodeCursor.next;
        }

    }
    
    /**
     * Calculates the depth of the nodes. Depth of a node shows how many arcs away the node is from the root node in
     * the tree structure. This method assumes that the <code>forwardThread</code> and <code>treePredecessor</code>
     * tree_algorithm_node_information_fileds are correct.
     */
    public final void calculateDepth(){
        Node rootNode;
        Node cursor;
        rootNode=findRoot();
        getAlgNodeInfo(rootNode).depth=0;
        cursor=getAlgNodeInfo(rootNode).forwardThread;
        while (cursor!=rootNode){
            getAlgNodeInfo(cursor).depth=getAlgNodeInfo( getAlgNodeInfo(cursor).treePredecessor ).depth+1;
            cursor=getAlgNodeInfo(cursor).forwardThread;
        }
    }
    
    /**
     * Sets up the <code>forwardThread</code> and <code>reverseThread</code> tree_algorithm_node_information_fileds for
     * all the nodes. This method assumes that <code>treePredecessor</code> fields are correct. Threads are extremely usefull.
     * They are simply pointers from each node to another node, in such a way that if you follow the threads starting from
     * the root node, you will come back to root node and in the process you will go thru every node, and the order you
     * go thru every node is such that you visit the <code>treePredecessor</code> of a node before you visit that node. So
     * when you follow the threads and you arive at a node, you can be sure that <code>treePredecessor</code> of this node
     * was visited before.
     */
    public final void buildThread(){
        Node rootNode;
        Node parent;
        Node child;
        Node nextNode;
        NodeList nodeCursor=targetNetwork.getNodeList();
        NodeList rootNodeInList=null;  
        Node myNode;
        
        rootNode=null;
        while (nodeCursor!=null) {
            myNode=nodeCursor.elem;
            if (getAlgNodeInfo(myNode).treePredecessor==null){
                rootNode=myNode;
                rootNodeInList=nodeCursor;
            }
            getAlgNodeInfo(myNode).forwardThread=null;
            nodeCursor=nodeCursor.next;
        }
        
        getAlgNodeInfo(rootNode).forwardThread=rootNode;
        getAlgNodeInfo(rootNode).reverseThread=rootNode;
        
        nodeCursor=rootNodeInList;
        nodeCursor=nodeCursor.next;
        if (nodeCursor==null) {nodeCursor=targetNetwork.getNodeList();}
        while (!nodeCursor.equals(rootNodeInList)) {
            myNode=nodeCursor.elem;
            if (getAlgNodeInfo(myNode).forwardThread==null) {
                child=nodeCursor.elem;
                parent=getAlgNodeInfo(child).treePredecessor;
                while (getAlgNodeInfo(parent).forwardThread==null) {
                    getAlgNodeInfo(parent).forwardThread=child;
                    getAlgNodeInfo(child).reverseThread=parent;
                    child=parent;
                    parent=getAlgNodeInfo(parent).treePredecessor;
                }
                nextNode=getAlgNodeInfo(parent).forwardThread;
                getAlgNodeInfo(myNode).forwardThread=nextNode;
                getAlgNodeInfo(nextNode).reverseThread=myNode;
                getAlgNodeInfo(parent).forwardThread=child;
                getAlgNodeInfo(child).reverseThread=parent;
            }
            nodeCursor=nodeCursor.next;
            if (nodeCursor==null) {nodeCursor=targetNetwork.getNodeList();}
        }
    }
    
    /**
     * Checks whether the tree struncture has many roots or the tree cycles etc.
     * @exception   TreeHasManyRootsException
     * @exception   TreeHasNoRootException
     * @exception   InvalidTreeException
     * @exception   CyclicTreeException
     */
    public final void checkTree() throws TreeHasManyRootsException, TreeHasNoRootException, InvalidTreeException, CyclicTreeException {
        NodeList nodeCursor=targetNetwork.getNodeList();
        int count;                                                                            /*count*/
        Node treeNode;
        Node rootNode=null;
        Node temporaryNode;
    
        while (nodeCursor!=null) {
            temporaryNode= nodeCursor.elem;
            getAlgNodeInfo(temporaryNode).flag=0;
            if (getAlgNodeInfo(temporaryNode).treePredecessor==null) {
                if (rootNode==null) {
                    rootNode=temporaryNode;
                }
                else {
                    throw new TreeHasManyRootsException("bad tree: tree has more than one root",rootNode,temporaryNode);
                }
            }
            nodeCursor=nodeCursor.next;
        }
    
        if (rootNode==null) {throw new TreeHasNoRootException("bad tree: tree has no root");}
    
        if (getAlgNodeInfo(rootNode).predecessorLink != null) {throw new InvalidTreeException("bad tree: tree predecessor structure is bad",rootNode);}
    
        nodeCursor=targetNetwork.getNodeList();
        getAlgNodeInfo(rootNode).flag=1;;
        count=2;
        while (nodeCursor!=null) {
            treeNode=nodeCursor.elem;
            while (getAlgNodeInfo(treeNode).flag!=1) {
                if  ((getAlgNodeInfo(treeNode).predecessorLink==null) ||    
                    (!(((getAlgNodeInfo(treeNode).predecessorLink.getTailNode()==treeNode) &&
                    (getAlgNodeInfo(treeNode).predecessorLink.getHeadNode()==getAlgNodeInfo(treeNode).treePredecessor)) ||
                    ((getAlgNodeInfo(treeNode).predecessorLink.getHeadNode()==treeNode) &&
                    (getAlgNodeInfo(treeNode).predecessorLink.getTailNode()==getAlgNodeInfo(treeNode).treePredecessor))))) {
                    
                        throw new InvalidTreeException("bad tree: tree predecessor structure is bad",treeNode);
                    }
                    else {
                        if (getAlgNodeInfo(treeNode).flag==count) {
                            throw new CyclicTreeException("bad tree: tree includes cycles",treeNode);
                        }
                        if (treeNode!=rootNode) getAlgNodeInfo(treeNode).flag=count;
                    }
                    treeNode=getAlgNodeInfo(treeNode).treePredecessor;
            }
            count++;
            nodeCursor=nodeCursor.next;
        }
    }
    
    /**
     * Checks the thread structure and returns the number of nodes traced by following 
     * the thread structure until it gets back to the root node
     */
    
    public int checkThread() throws InvalidTreeException
    {
         NodeList nodeList = targetNetwork.getNodeList();
         int count = 1;
         
         while ( nodeList != null )
         {
            getAlgNodeInfo( nodeList.elem ).flag = 0;
            nodeList = nodeList.next;
         }
         
         Node rootNode = findRoot();
         getAlgNodeInfo( rootNode ).flag = 1;
         Node currentNode = getAlgNodeInfo( rootNode ).forwardThread;
         
         while ( ! ( rootNode.equals ( currentNode ) ) )
         {
            if ( getAlgNodeInfo( getAlgNodeInfo( currentNode ).treePredecessor ).flag == 0 )
            {
               throw new InvalidTreeException ( "invalid thread" , currentNode );
            }
            
            getAlgNodeInfo( currentNode ).flag = 1;
            count ++ ;
            currentNode = getAlgNodeInfo( currentNode ).forwardThread;
         }
         return count;
    }
    
    
    /**
     * Returns an array of arc that are in the tree.
     * @exception   TreeHasManyRootsException
     * @exception   TreeHasNoRootException
     * @exception   InvalidTreeException
     * @exception   CyclicTreeException
     */
    public final Arc[] getTree() throws TreeHasManyRootsException, TreeHasNoRootException, InvalidTreeException, CyclicTreeException {
       
        checkTree();
        Node rootNode=findRoot();
        buildThread();
        Arc[] treeArcs=new Arc[targetNetwork.getNoOfNodes()-1];
        Node nodeCursor=getAlgNodeInfo(rootNode).forwardThread;
        int i=0;
        while ( nodeCursor!=rootNode) {
            treeArcs[i]=getAlgNodeInfo(nodeCursor).predecessorLink;
            i++;
            nodeCursor=getAlgNodeInfo(nodeCursor).forwardThread;
        }
        return treeArcs;
    }
    
    protected final void initializeNodeFlags(int initializeTo){
        
        Node rootNode=findRoot();
        Node nodeCursor=getAlgNodeInfo(rootNode).forwardThread;
        buildThread();
        while (nodeCursor!=rootNode) {
            getAlgNodeInfo(nodeCursor).flag=initializeTo;
            nodeCursor=getAlgNodeInfo(nodeCursor).forwardThread;
        }
    }
    
    public String toString() {
        return super.toString();
    }
    
    


}



    
    
    
    
    