package Network.Representation;

import Network.Representation.*;

/**
 * This class in fact simulates a "retailer" rather than a "factory", because if it runs out of
 * nodes, it returns null rather than producing a new node. In this way we were able to code "node factories"
 * rather than "node implementation factories". So users can store anything that implements <code>Node</code> in
 * this factory and retrieve it when necessary.
 */
 
public class NodeFactory {
    
    private NodeStack storedNodes;
    private Network anchorNetwork;
    
    /**
     * Creates an node factory with given initial size and size increment.
     */
    public NodeFactory (int initialSize, int sizeIncrement) {
        storedNodes=new NodeStack (initialSize, sizeIncrement);
        anchorNetwork=new NetworkImplementation();
    }
    
    /** 
     * Get a node from the inventory of node factory. If the factory has run out nodes, this method returns null. Then
     * the user is free to produce any kind of node he wants instead of getting a default node class that would otherwise
     * be used by the factory.
     *
     * @param   nodeSupply the supply of the node to be given by the node factory.
     */
     
    public Node makeNode(int nodeSupply){
        Node dummyNode;
        dummyNode=storedNodes.pop();
        if ( dummyNode!=null ) {
            dummyNode.initializeNode(nodeSupply);
        }
        return (dummyNode);
    }
    
    
    /**
     * Puts the given node ino the inventory of the node factory. This node should not belong to any network and should
     * not point to any other networks. Otherwise <code>NodeError</code> will be thrown.
     */
     
    public void collectNode(Node collectedNode) {
        storedNodes.push(collectedNode,anchorNetwork);
    }
    
    public String toString() {
        return storedNodes.toString();
    }

        
}





class NodeStack {
    
    Node[] elements;
    int capacityIncrement;
    int noOfNodesStored;
    
    
    NodeStack (int size, int sizeIncrement) {
        elements=new Node[size];
        capacityIncrement=sizeIncrement;
        noOfNodesStored=0;
    }
    
    final synchronized void push (Node pushedNode, Network flagNetwork){
        if (pushedNode.getContainerNetwork()==null) {
            
            if ( noOfNodesStored >= elements.length ) {
                resizeElements();
            }
            elements[noOfNodesStored]=pushedNode;
            pushedNode.setContainerNetwork(flagNetwork);
            noOfNodesStored++;
        }
        else {
            throw new NodeError("node can't be pushed into node factory: node already belongs to a network");
        }
    }
    
    
    
    final synchronized Node pop() {
        
        if (noOfNodesStored > 0 ) {
            Node dummyNode=elements[--noOfNodesStored];
            dummyNode.setContainerNetwork(null);
            return dummyNode;
        }
        else {
            return null;
        }
    }
    
    final synchronized boolean isEmpty() {
        return (noOfNodesStored==0);
    }
        
    
    final synchronized void resizeElements(){
        Node[] newElements=new Node[elements.length+capacityIncrement];
        System.arraycopy (elements, 0, newElements, 0, elements.length);
        elements=newElements;
    }
    
    public String toString() {
        return Integer.toString(noOfNodesStored)+" "+Integer.toString(elements.length)+" "+Integer.toString(capacityIncrement);
    }
    
    
}    

